
/**
 * @description
 * @typedef {Object} VecRenderNumberVertices
 * @property {number[]} lineVertices
 * @property {number[]} maskVertices
 */

/**
 * @description
 * @typedef {Object} WaveValueFormatMaxMin
 * @property {number} maxVal
 * @property {number} minVal
 */

import { maskColorIndexOffset } from "../render-utils";
import { hexToSignedInt } from "./renderFormat";


function makeQuadVertices(p1, p2, p3, p4) {
    return [
        ...p1, ...p2, ...p3,
        ...p1, ...p3, ...p4
    ]
}

/**
 * @description Digital
 * @param {GlobalLookup} lookup
 * @param {string} link
 * @param {(string | number)[]} wave
 * @param {number} time
 * @returns {VecRenderNumberVertices} 
 */
export function renderAsCommonDigital(lookup, link, wave, time) {
    const length = wave.length;
    const lineVertices = [];
    const maskVertices = [];
    
    for (let i = 0; i < length; ++ i) {
        const [t1, val, mask] = wave[i];
        const t2 = (i === (length - 1)) ? time : wave[i + 1][0];

        // t1(val) --- t2(val)
        // 详见 设计图 ./design/webgl.drawio makeVecVertex原理 sheet
        // 此处的 y 不是 y 的索引，详见 gl_Shifts_for_bar 的设计
        const p0 = {x: t1, y: 0};
        const p1 = {x: t2, y: 0};
        const a0 = {x: t1, y: 4};
        const a1 = {x: t2, y: 5};
        const a2 = {x: t1, y: 1};
        const a3 = {x: t2, y: 2};

        let color = 5;
        const renderOptions = lookup.currentSignalRenderOptions;
        if (mask) {
            color = 4;
        } else if (renderOptions.has(link)) {
            const option = renderOptions.get(link);
            if (typeof option.color === 'number') {
                color = option.color;
            }
        }

        const points = [ a1, p1, a3, a2, p0, a0 ];
        const wsIndice = [ 1, 2, 3, 5, 6, 7 ];
        for (let j = 0; j < 6; ++ j) {
            const prePoint = points[j];
            const nextPoint = points[(j + 1) % 6];
            const preWsIndex = wsIndice[j];
            const nextWsIndex = wsIndice[(j + 1) % 6];
            const quadVertices = makeQuadVertices( 
                [prePoint.x,  prePoint.y,  color, preWsIndex],
                [prePoint.x,  prePoint.y,  color, (preWsIndex + 4) % 8],
                [nextPoint.x, nextPoint.y, color, (nextWsIndex + 4) % 8],
                [nextPoint.x, nextPoint.y, color, nextWsIndex],
            );
            lineVertices.push(...quadVertices);
        }

        const maskTriangles = [
            [3, 4, 5],
            [5, 0, 3],
            [0, 3, 2],
            [0, 1, 2]
        ];
        
        for (const triangleIndice of maskTriangles) {
            const [i1, i2, i3] = triangleIndice;
            const point1 = points[i1];
            const point2 = points[i2];
            const point3 = points[i3];
            maskVertices.push(
                point1.x, point1.y, color + maskColorIndexOffset, wsIndice[i1],
                point2.x, point2.y, color + maskColorIndexOffset, wsIndice[i2],
                point3.x, point3.y, color + maskColorIndexOffset, wsIndice[i3]
            );
        }
    }

    return { lineVertices, maskVertices };
}


/**
 * 
 * @param {GlobalLookup} lookup 
 * @param {string} link 
 */
function getValFormatCode(lookup, link) {
    const renderOptions = lookup.currentSignalRenderOptions;
    if (renderOptions.has(link)) {
        const option = renderOptions.get(link);
        if (typeof option.valueFormat === 'number') {
            return option.valueFormat;
        } else {
            return 0;
        }
    }

    return 0;
}

/**
 * @description 计算为 js 的数字，如果当前模式为 有符号整型，需要先转换成 有符号数进行比较
 * @param {number} formatCode 
 * @param {number | string} val 
 * @param {number} width 
 */
function explainAsJSNumber(formatCode, val, width) {
    if (formatCode === 3) {
        if (typeof val === 'number') {
            val = val.toString(16);
        }
        if (val.startsWith('0x')) {
            val = val.substring(2);
        }
        if (val.length < width) {
            val = '0'.repeat(width - val.length) + val;
        }
        return hexToSignedInt(val);
    } else {
        return parseInt(val);
    }
}

/**
 * @description Analog (Ladder) 阶梯状
 * @param {string} link
 * @param {(string | number)[]} wave
 * @param {number} time
 * @param {number} formatCode
 * @param {number} width
 * @returns {WaveValueFormatMaxMin} 
 */
function getMaxMinByFormat(link, wave, time, formatCode, width) {
    let maxVal = undefined;
    let minVal = undefined;

    // TODO : 优化这段代码
    for (let i = 0; i < wave.length; ++ i) {
        const [t1, val, mask] = wave[i];
        const t2 = (i === (length - 1)) ? time : wave[i + 1][0];
        const numVal = explainAsJSNumber(formatCode, val, width);

        if (maxVal === undefined) {
            maxVal = numVal;
        } else {
            maxVal = Math.max(maxVal, numVal);
        }

        if (minVal === undefined) {
            minVal = numVal;
        } else {
            minVal = Math.min(minVal, numVal);
        }
    }

    return { maxVal, minVal };
}

/**
 * 
 * @param {number} formatCode 
 * @param {number} maxVal 
 * @param {number} minVal 
 * @returns {(val: number) => number}
 */
function getMappingFunc(formatCode, maxVal, minVal) {
    if (formatCode === 3) {
        const maxAbs = Math.max(Math.abs(maxVal), Math.abs(minVal));
        const k = 1 / maxAbs;
        const b = -1;
        // 把 [-maxAbs, +maxAbs] 映射到 [-1, 1]
        return val => k * (val + maxAbs) + b;
    } else {
        const k = 2 / maxVal;
        const b = -1;
        // 把 [0, max] 映射到 [-1, 1]
        return val => k * val + b;
    }
}

/**
 * @description Analog (Ladder) 阶梯状
 * @param {GlobalLookup} lookup
 * @param {string} link
 * @param {(string | number)[]} wave
 * @param {number} time
 * @returns {VecRenderNumberVertices} 
 */
export function renderAsLadderAnalog(lookup, link, wave, time) {
    const renderOptions = lookup.currentSignalRenderOptions;
    const formatCode = getValFormatCode(lookup, link);
    const signal = lookup.link2CurrentWires.get(link);
    const width = signal.size;

    const { maxVal, minVal } = getMaxMinByFormat(lookup, link, wave, time, formatCode, width);
    const coordinateTransform = getMappingFunc(formatCode, maxVal, minVal);

    const length = wave.length;
    const lineVertices = [];
    const maskVertices = [];

    for (let i = 0; i < length; ++ i) {
        const [t1, val, mask] = wave[i];
        const t2 = (i === (length - 1)) ? time : wave[i + 1][0];

        let color = 5;
        if (mask) {
            color = 4;
        } else if (renderOptions.has(link)) {
            const option = renderOptions.get(link);
            if (typeof option.color === 'number') {
                color = option.color;
            }
        }

        

        const numVal = explainAsJSNumber(formatCode, val, width);
        const y = coordinateTransform(numVal);
        
    }
}

/**
 * @description Analog (Line) 模拟状
 * @param {GlobalLookup} lookup
 * @param {string} link
 * @param {(string | number)[]} wave
 * @param {number} time
 * @returns {VecRenderNumberVertices} 
 */
export function renderAsLineAnalog(lookup, link, wave, time) {
    const renderOptions = lookup.currentSignalRenderOptions;
    const formatCode = getValFormatCode(lookup, link);

    const { maxVal, minVal } = getMaxMinByFormat(lookup, link, wave, time, formatCode);
    
}